/*
  RA decompression class version 1.0
  (c) 2006 Benjamin Haisch
  email: john_doe@techie.com
  You may use this code freely as long as you give me credit for it :)
*/

#include "decompressRA.h"

int DecompressRA::getBit() {
    if (bitsLeft == 0) {
        bitBuffer = *inb++;
        bitsLeft = 8;
    }
    byte t = (bitBuffer & 0x80) >> 7;
    bitBuffer <<= 1;
    bitsLeft--;
    return t;
}

int DecompressRA::loadBits(int count) {
    int value = 0;
    while (count > 0) {
        value = (value << 1) | getBit();
        count--;
    }
    return value;
}

void DecompressRA::loadTable(byte *table1, byte *table2) {
    int count = (*inb++) * 2;
    int len1 = loadBits(3) + 1;
    int len2 = loadBits(3) + 1;
    while (count > 0) {
        byte value;
        if (getBit() == 0) {
            value = loadBits(len1);
        } else {
            value = loadBits(len2);
            *table2 = value;
            value = 0;
        }
        *table1++ = value;
        table2++;
        count--;
    }
}

int DecompressRA::loadValue(byte *table1, byte *table2) {
    int index = 0;
    byte value = 0;
    do {
        index = (value << 1) | getBit();
        value = table1[index];
    } while (value != 0);
    return table2[index];
}

int DecompressRA::loadCount() {
    return loadValue(countTable1, countTable2);
}

int DecompressRA::loadLength() {
    int length = loadValue(lengthTable1, lengthTable2);
    if (length == 255) {
        byte len_hi = (getBit() << 1) | getBit();
        byte len_lo = *inb++;
        length = ((len_hi << 8) | len_lo) + 255;
    }
    return length;
}

int DecompressRA::loadOffset() {
    int offset = loadValue(offsetTable1, offsetTable2);
    if (offset >= 2) {
        int bitcount = offset - 1;
        offset = 1;
        while (bitcount > 0) {
            offset = (offset << 1) | getBit();
            bitcount--;
        }
    }
    return ~offset;
}

void DecompressRA::decompressData() {

    bitsLeft = 0;
    bitBuffer = 0;

    memset(countTable1, 0, 512);
    memset(countTable2, 0, 512);
    memset(lengthTable1, 0, 512);
    memset(lengthTable2, 0, 512);
    memset(offsetTable1, 0, 32);
    memset(offsetTable2, 0, 32);

    /* initialize bitbuffer */
    getBit();

    do {

        uint16 blocksize = *(uint16*)inb;
        inb += 2;

        /* load tables */
        loadTable(lengthTable1, lengthTable2);
        loadTable(countTable1, countTable2);
        loadTable(offsetTable1, offsetTable2);

        /* process the block */
        while (blocksize > 0) {
            int copyCount = loadCount();
            memcpy(outb, inb, copyCount);
            inb += copyCount;
            outb += copyCount;
            int length = loadLength();
            if (length > 0) {
                int offset = loadOffset();
                /* don't use memcpy here because source and dest can overlap */
                while (length > 0) {
                    *outb++ = *(outb+offset);
                    length--;
                }
            }
            blocksize--;
        }

    } while (getBit());

}

bool DecompressRA::isRA(byte *inbuffer, uint32 insize) {
    return (insize > 12 && *(uint32*)inbuffer == 0x02004152);
}

uint32 DecompressRA::decompress(byte *inbuffer, byte **outbuffer) {

    if (!outbuffer || !inbuffer)
        return 0;

    uint32 outsize = *(uint32*)&inbuffer[4];
    *outbuffer = new byte[outsize];

    /* set input to the beginning of the compressed data */
    inb = inbuffer + 12;
    outb = *outbuffer;
    decompressData();

    /* return the decompressed size */
    return outb - *outbuffer;

}
